from enum import Enum, unique

@unique
class CsclIndices(Enum):
    SCORE = 'scor'
    SOCIAL_KB = 'social_kb'
    NO_CONTRIBUTION = 'no_contribution'
    NO_NOUNS = 'no_nouns'
    NO_VERBS = 'no_verbs'
    OUTDEGREE = 'outdegree'
    INDEGREE = 'indegree'




from datetime import date

class CommunityProcessing:

	@staticmethod
	def determine_participant_contributions(community):
		for c in community.get_conversations():
			for p in c.get_participants():
				if p.get_contributions().get_blocks() != None and len(p.get_contributions().get_blocks()) != 0:
					index = community.get_participants().index(p)
					participant_to_update = None

					if index >= 0:
						participant_to_update = community.get_participants()[index]
					else:
						participant_to_update = Participant(p.get_name(), c)
						community.get_participants().append(participant_to_update)

					for block in p.get_contributions().get_blocks():
						contribution = Contribution(block)

						if contribution != None and contribution.get_time() != None and contribution.is_eligible(community.get_start_date(), community.get_end_date()):
							if community.get_first_contribution_date() != None:
								community.set_first_contribution_date(contribution.get_time())

							if contribution.get_time() < community.get_first_contribution_date():
								community.set_first_contribution_date(contribution.get_time())

								# gregorian calendar 2010
								date_base = date(2010, 1, 1)

							if community.get_last_contribution_date() is None:
								community.set_last_contribution_date(contribution.get_time())

							if contribution.get_time() > community.get_last_contribution_date():
								community.set_last_contribution_date(contribution.get_time())

							Block.add_block(participant_to_update.get_contributions(), block)
							Block.add_block(community.get_eligible_contributions(), block)
							if block.is_significant():
								Block.add_block(participant_to_update.get_significant_contributions(), block)

							indices = participant_to_update.get_indices()
							indices[CSCLIndices.NO_CONTRIBUTION] += 1

							for key, value in contribution.get_word_occurences().items():
								if key.get_POS() != None:
									if key.get_POS().startswith("N"):
										indices[CSCLIndices.NO_NOUNS] += value

									if key.get_POS().startswith("V"):
										indices[CSCLIndices.NO_VERBS] += value

		if community.get_start_date() is None:
			community.set_start_date(community.get_first_contribution_date())
		if community.get_end_date() is None:
			community.set_end_date(community.get_last_contribution_date())

	@staticmethod
	def determine_participation(community):
		size = len(community.get_participants())
		community.set_participant_contributions(size, size)

		for c in community.get_conversations():
			blocks = c.get_blocks()

			for i in range(len(blocks)):
				contribution = Contribution(blocks[i])

				if contribution != None and contribution.get_time() != None and contribution.is_eligible(community.get_start_date(), community.get_end_date()):
					p1 = contribution.get_participant()
					index1 = community.get_participants().index(p1)

					if index1 >= 0:
						community.get_participant_contributions()[index1][index1] += contribution.get_score()
						participant_to_update = community.get_participants()[index1]

						indices = participant_to_update.get_indices()
						indices[CSCLIndices.SCORE] += contribution.get_score()
						indices[CSCLIndices.SOCIAL_KB] += contribution.get_social_KB()

						for j in range(i):
							if c.get_prunned_block_distances()[i][j] != None:
								p2 = (Contribution(blocks[j])).get_participant()
								index2 = community.get_participants().index(p2)

								if index2 >= 0:
									added_KB = blocks[i].get_score() * c.get_prunned_block_distances()[i][j].get_cohesion()
									community.get_participant_contributions()[index1][index2] += added_KB

	@staticmethod
	def compute_SNA_metrics(community):
		ParticipantEvaluation.perform_SNA_participants(community.get_participants(), community.get_participant_contributions(), True)

		for c in community.get_conversations():
			p = None
			blocks = c.get_blocks()

			for i in range(len(blocks)):
				if blocks[i] != None:
					if p is None:
						p = (Contribution(blocks[i])).get_participant()
						participant_to_update = community.get_participants()[community.get_participants().index(p)]
						indices = participant_to_update.get_indices()

						indices[CSCLIndices.NO_NEW_THREADS] += 1
						indices[CSCLIndices.NEW_THREADS_OVERALL_SCORE] += c.get_score()
						indices[CSCLIndices.NEW_THREADS_CUMULATIVE_SOCIAL_KB] +=
							VectorAlgebra.sum_elements(c.get_social_KB_evolution())
						indices[CSCLIndices.AVERAGE_LENGTH_NEW_THREADS] +=
							len(blocks[i].get_text())

						break

		filtered_participants = list(filter(lambda p: (p.get_indices()[CSCLIndices.NO_NEW_THREADS] != 0),
									community.get_participants()))

		for p in filtered_participants:
			indices = p.get_indices()
			indices[CSCLIndices.AVERAGE_LENGTH_NEW_THREADS] /= indices[CSCLIndices.NO_NEW_THREADS]


	



from typing import List, Dict, Tuple

from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.core.text_element_type import TextElementType
from rb.core.cscl.contribution import Contribution
from rb.core.cscl.conversation import Conversation
from rb.core.cscl.participant import Participant
from rb.cna.cna_graph import CnaGraph
from rb.similarity.vector_model_factory import create_vector_model
from rb.similarity.vector_model import VectorModelType

from rb.utils.rblogger import Logger

logger = Logger.get_logger()

# muchii intre replici parinte-copil?
# muchii intre replici ale aceluiasi user ?

class Community(TextElement):

	'''
		community -> list of JSON objects (crawled from Reddit), with
					 the same structure as described in Conversation class
	'''
	def __init__(self, lang: Lang, 
                 community: List[Dict],
				 container: TextElement = None,
                 depth: int = TextElementType.COMM.value,
                 start_date: int = None,
                 end_date: int = None):

		TextElement.__init__(self, lang=lang, text=None,
                             depth=depth, container=container)

		self.components = []
		self.start_date = start_date
		self.end_date = end_date

		self.participant_map = dict()

		for conversation_thread in community:
			current_conversation = Conversation(lang, container=self,
						 						conversation_thread=conversation_thread)

			self.components.append(current_conversation)

		self.participants = self.union_participants()
		self.participant_contributions = self.union_contributions()
		self.first_contribution_date, self.last_contribution_date = self.find_contribution_range()


	def union_participants(self) -> List[Participant]:
		return [self.participant_map[participant_id] for participant_id in self.participant_map]

	def union_contributions(self) -> Dict:
		contributions = dict()

		for conversation in self.components:
			for participant in conversation.get_participants():
				if not (participant in contributions):
					contributions[participant.get_id()] = []

				contributions[participant.get_id()] += conversation.get_participant_contributions(participant.get_id())

		return contributions

	def find_contribution_range(self) -> Tuple[int, int]:
		first_contribution = None
		last_contribution = None

		for conversation in self.components:
			for contribution in conversation.get_contributions():
				timestamp = contribution.get_timestamp()

				if first_contribution == None or timestamp < first_contribution:
					first_contribution = timestamp

				if last_contribution == None or timestamp > last_contribution:
					last_contribution = timestamp

		return first_contribution, last_contribution

	def get_conversations(self) -> List[Conversation]:
		return self.components

	def get_participants(self) -> List[Participant]:
		return self.participants

	def get_participant_contributions(self, participant_id: str) -> List[Contribution]:
		return self.participant_contributions[participant_id]

	def get_first_contribution_date(self) -> int:
		return self.first_contribution_date

	def get_last_contribution_date(self) -> int:
		return self.last_contribution_date



import csv
import math
import os
from heapq import heapify, heappop
from typing import List, Dict, Iterable

from rb.cna.cna_graph import CnaGraph
from rb.complexity.complexity_index import compute_indices, ComplexityIndex
from rb.core.document import Document
from rb.core.lang import Lang
from rb.processings.pipeline.dataset import Dataset, TargetType, Task
from rb.similarity.vector_model_factory import get_default_model
from scipy.stats import f_oneway, pearsonr
from rb.processings.pipeline.estimator import Estimator
from rb.processings.pipeline.svm import SVM
from rb.processings.pipeline.svr import SVR
from rb.processings.pipeline.ridge_regression import RidgeRegression
from rb.processings.pipeline.mlp import MLP
from rb.processings.pipeline.random_forrest import RandomForest

CLASSIFIERS = [SVM, RandomForest, MLP]
REGRESSORS = [SVR, RidgeRegression]

def construct_documents(dataset: List[str], lang: Lang) -> List[Document]:
    model = get_default_model(lang)
    result = []
    for text in dataset:
        doc = Document(lang, text)
        cna_graph = CnaGraph(docs=doc, models=[model])
        compute_indices(doc=doc, cna_graph=cna_graph)
        result.append(doc)
    return result

def filter_rare(dataset: Dataset):
    features = []
    for index in dataset.features:
        values = [doc.indices[index] for doc in dataset.train_docs]
        zeros = sum(1 for val in values if val == 0)
        if zeros / len(values) < 0.2:
            features.append(index)
    dataset.features = features

def preprocess(folder: str, targets_file: str, limit: int = None) -> Dataset:
    files = {filename.replace(".txt", "").strip(): open(os.path.join(folder, filename), "rt", encoding='utf-8', errors='ignore').read().strip() 
             for filename in os.listdir(folder)
             if not filename.startswith(".")}
    texts = []
    targets = []
    with open(targets_file, "rt", encoding='utf-8') as f:
        reader = csv.reader(f, delimiter=',')
        header = next(reader)
        for line in reader:
            filename = line[0].replace(".txt", "").strip()
            if filename not in files:
                print(filename)
                continue
            texts.append(files[filename])
            targets.append(line[1:])
    dataset = Dataset(texts, targets)
    if limit is not None:
        dataset.train_texts = dataset.train_texts[:limit]
        dataset.dev_texts = dataset.dev_texts[:limit]
        for task in dataset.tasks:
            task.train_values = task.train_values[:limit]
            task.dev_values = task.dev_values[:limit]
    
    dataset.train_docs = construct_documents(dataset.train_texts, Lang.EN)
    dataset.dev_docs = construct_documents(dataset.dev_texts, Lang.EN)
    dataset.features = list(dataset.train_docs[0].indices.keys())
    filter_rare(dataset)
    for task in dataset.tasks:
        remove_colinear(dataset, task)
    return dataset
    # dataset.save_features("features.csv")

def correlation_with_targets(feature: ComplexityIndex, dataset: Dataset, task: Task) -> float:
    values = [doc.indices[feature] for doc in dataset.train_docs]
    if task.type is TargetType.FLOAT:
        corr, p = pearsonr(values, task.train_values)
        return abs(corr)
    values_per_class = {}
    for val, target in zip(values, task.train_values):
        if target not in values_per_class:
            values_per_class[target] = []
        values_per_class[target].append(val)
    return f_oneway(*values_per_class.values())

def remove_colinear(dataset: Dataset, task: Task) -> None:
    # features = list(zip(*Dataset.load_features("features.csv")))
    heap = []
    for i, a in enumerate(dataset.features[:-1]):
        for j, b in enumerate(dataset.features[i+1:]):
            values_a = [doc.indices[a] for doc in dataset.train_docs]
            values_b = [doc.indices[b] for doc in dataset.train_docs]
            corr, p = pearsonr(values_a, values_b)
            if math.isnan(corr):
                continue
            heap.append((-corr, i, j))
    heapify(heap)
    
    correlations = [correlation_with_targets(feature, dataset, task) 
                    for feature in dataset.features]
    task.mask = [True] * len(dataset.features)
    while len(heap) > 0:
        inv_corr, i, j = heappop(heap)
        if not task.mask[i] or not task.mask[j]:
            continue
        if inv_corr < -0.9:
            if correlations[i] > correlations[j]:
                task.mask[j] = False
            else:
                task.mask[i] = False


def next_config(parameters: Dict[str, List]) -> Iterable[Dict[str, str]]:
    solution  = [-1] * len(parameters)
    keys = list(parameters.keys())
    current = 0
    while current >= 0:
        if current == len(parameters):
            yield {keys[i]: parameters[keys[i]][j] for i, j in enumerate(solution)}
            current -= 1
        if solution[current] == len(parameters[keys[current]]) - 1:
            solution[current] = -1
            current -= 1
            continue
        solution[current] += 1
        current += 1
            

def grid_search(dataset: Dataset, task: Task) -> Estimator:
    if task.type is TargetType.FLOAT:
        estimators = REGRESSORS
    else:
        estimators = CLASSIFIERS
    for estimator in estimators:
        parameters = estimator.parameters()
        for config in next_config(parameters):
            model = estimator(dataset, [task], config)
            acc = model.cross_validation()
            print("{} - {}".format(estimator, acc))
            
           
    


from typing import List, Dict
import csv
from dateutil import parser
from datetime import datetime
import xmltodict

from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.core.text_element_type import TextElementType
from rb.core.cscl.contribution import Contribution
from rb.core.cscl.conversation import Conversation
from rb.core.cscl.community import Community
from rb.core.cscl.cscl_indices import CsclIndices
from rb.cna.cna_graph import CnaGraph
from rb.similarity.vector_model_factory import create_vector_model
from rb.similarity.vector_model import VectorModelType
from rb.processings.cscl.participant_evaluation import ParticipantEvaluation

from rb.utils.rblogger import Logger

logger = Logger.get_logger()

CONTRIBUTIONS_KEY = 'contributions'
ID_KEY = 'id'
PARENT_ID_KEY = 'parent_id'
TIMESTAMP_KEY = 'timestamp'
TEXT_KEY = 'text'
USER_KEY = 'user'

class CsvParser:

	@staticmethod
	def get_json_from_csv(filename: str) -> Dict:
		conversation_thread = dict()
		contribution_list = []

		with open(filename) as csv_file:
			csv_reader = csv.reader(csv_file, delimiter=',')
			first = True			

			for row in csv_reader:
				if first:
					first = False
				else:
					contribution = dict()

					contribution[ID_KEY] = row[0]
					contribution[PARENT_ID_KEY] = row[1]
					contribution[USER_KEY] = row[2]
					contribution[TEXT_KEY] = row[3]
					contribution[TIMESTAMP_KEY] = row[4]

					contribution_list.append(contribution)

		conversation_thread[CONTRIBUTIONS_KEY] = contribution_list

		return conversation_thread

	@staticmethod
	def parse_large_csv(filename: str) -> Dict:
		conversation_thread = dict()
		contribution_list = []

		with open(filename) as csv_file:
			csv_reader = csv.reader(csv_file, delimiter=',')
			first = True			

			for row in csv_reader:
				if first:
					first = False
				else:
					contribution = dict()

					contribution[ID_KEY] = row[0]
					contribution[PARENT_ID_KEY] = row[1]
					if row[1] == '':
						contribution[PARENT_ID_KEY] = '-1'

					contribution[USER_KEY] = row[2]
					contribution[TEXT_KEY] = row[6]

					contribution[TIMESTAMP_KEY] = datetime.timestamp(parser.parse(row[3], ignoretz=True, fuzzy=True))

					contribution_list.append(contribution)

		conversation_thread[CONTRIBUTIONS_KEY] = contribution_list

		return conversation_thread

	@staticmethod
	def load_from_xml(lang: Lang, filename: str) -> Dict:
		with open(filename, "rt") as f:
			my_dict=xmltodict.parse(f.read())
			contributions = [
				{
					ID_KEY: int(utterance["@genid"]) - 1,
					PARENT_ID_KEY: int(utterance["@ref"]) - 1,
					TIMESTAMP_KEY: datetime.timestamp(datetime.strptime(utterance["@time"],'%H.%M.%S').replace(year=2020)),
					USER_KEY: turn["@nickname"],
					TEXT_KEY: utterance["#text"],
				}
				for turn in my_dict["corpus"]["Dialog"]["Body"]["Turn"]
				for utterance in (turn["Utterance"] if isinstance(turn["Utterance"], List) else [turn["Utterance"]])
			]

			return {CONTRIBUTIONS_KEY: contributions}

def compute_indices(conv: Conversation):
	participant_list = conv.get_participants()
	names = list(map(lambda p: p.get_id(), participant_list))

	print('Participants are:')
	print(names)

	print('Begin computing indices')

	ParticipantEvaluation.evaluate_interaction(conv)
	ParticipantEvaluation.evaluate_involvement(conv)
	ParticipantEvaluation.evaluate_used_concepts(conv)
	ParticipantEvaluation.perform_sna(conv, False)

	print('Finished computing indices')

	for p in participant_list:
		print('Printing for participant ' + p.get_id())

		print(p.get_index(CsclIndices.SCORE))
		print(p.get_index(CsclIndices.NO_NOUNS))
		print(p.get_index(CsclIndices.NO_VERBS))
		print(p.get_index(CsclIndices.NO_CONTRIBUTION))
		print(p.get_index(CsclIndices.SOCIAL_KB))
		print(p.get_index(CsclIndices.INDEGREE))
		print(p.get_index(CsclIndices.OUTDEGREE))

		print('---------------------')

	for n1 in names:
		for n2 in names:
			print('Score for ' + n1 + ' ' + n2 + ' is:')
			print(conv.get_score(n1, n2))

def main():
	# Test for English discussion CSV

	print('Testing English CSV')

	conv_thread = CsvParser.parse_large_csv('./thread.csv')

	community = Community(lang=Lang.EN, container=None, community=[conv_thread])
	en_coca_word2vec = create_vector_model(Lang.EN, VectorModelType.from_str("word2vec"), "coca")
	community.graph = CnaGraph(docs=[community], models=[en_coca_word2vec])

	conv = community.get_conversations()[0]

	compute_indices(conv)

	# Test for French discussion XML

	'''

	print('Testing French XML')

	conv_thread = CsvParser.load_from_xml(Lang.FR, './conpa-MEEF-anonyme.xml')
	
	community = Community(lang=Lang.FR, container=None, community=[conv_thread])
	fr_coca_word2vec = create_vector_model(Lang.FR, VectorModelType.from_str("word2vec"), "coca")
	community.graph = CnaGraph(docs=[community], models=[fr_coca_word2vec])

	conv = community.get_conversations()[0]

	compute_indices(conv)

	'''

if __name__ == '__main__':
	main()
